from typing import Dict, Callable, Any

from dataclasses import dataclass

from petisco.application.repository import Repository
from petisco.application.service import Service
from petisco.application.singleton import Singleton
from petisco.logger.logger import Logger, INFO, ERROR


@dataclass
class ApplicationConfig(metaclass=Singleton):
    services_provider: Callable[[], Dict[str, Service]]
    repositories_provider: Callable[[], Dict[str, Repository]]
    options: Dict[str, Any]
    info: Dict

    @staticmethod
    def get_instance():
        return ApplicationConfig()

    def __init__(
        self,
        mode: str,
        logger: Logger,
        config_dependencies: Callable = None,
        config_persistence: Callable = None,
        services_mode_mapper: Dict[str, Callable] = None,
        repositories_mode_mapper: Dict[str, Callable] = None,
        options: Dict[str, Any] = None
    ):
        """

        Parameters
        ----------
        mode
            DeploymentMode define the application mode of execution. If you're mapping services and repositories, please
            check given mode is mapped in services_mode_mapper and repositories_mode_mapper
        logger
            Pre configured logger
        config_dependencies
            Callable function to configure dependencies (e.g configure credentials in order to connect with a thrid-party
            application.
        config_persistence
            Callable function to configure application persistence (e.g configure a database)
        services_mode_mapper
            A dictionary to map DeploymentMode with a service provider function. This is used as a dependency injector
        repositories_mode_mapper
            A dictionary to map DeploymentMode with a repository provider function. This is used as a dependency injector
        """

        self.mode = mode
        self.logger = logger
        self.logger.log(INFO, "Deploying in {} mode".format(mode))

        if config_dependencies:
            config_dependencies()

        if config_persistence:
            config_persistence()

        self.info = {}
        if services_mode_mapper:
            if mode not in services_mode_mapper:
                self.logger.log(ERROR, f"Mode {mode} not found in services_mode_mapper ({services_mode_mapper})")
                raise NotImplementedError(f"Mode {mode} not found in services_mode_mapper ({services_mode_mapper})")
            self.services_provider = services_mode_mapper[mode]
            info_services = [(key, service.info()) for key, service in self.services_provider().items()]
            self.info["services"] = info_services

        if repositories_mode_mapper:
            if mode not in repositories_mode_mapper:
                self.logger.log(ERROR, f"Mode {mode} not found in repositories_mode_mapper ({repositories_mode_mapper})")
                raise NotImplementedError(f"Mode {mode} not found in repositories_mode_mapper ({repositories_mode_mapper})")
            self.repositories_provider = repositories_mode_mapper[mode]
            info_repositories = [(key, repository.info()) for key, repository in self.repositories_provider().items()]
            self.info["repositories"] = info_repositories

        self.options = options



# coding=utf-8
# Copyright (C) 2019+ Alice, Vigo, Spain

"""Public API of ALiCE Petisco Framework"""


# Classes
from petisco.application.application_config import ApplicationConfig
from petisco.application.repository import Repository
from petisco.application.service import Service
from petisco.use_case import UseCase
from petisco.use_case import use_case_logger
from petisco.controller.controller_decorator import controller
from petisco.controller.correlation_id import CorrelationId
from petisco.controller.errors.http_error import HttpError
classes = ["ApplicationConfig", "Service", "Repository", "UseCase", "use_case_logger", "controller", "CorrelationId", "HttpError"]

# Constants
from petisco.logger.logger import CRITICAL, FATAL, ERROR, WARNING, WARN, INFO, DEBUG, NOTSET
constants = ["CRITICAL", "FATAL", "ERROR", "WARNING", "WARN", "INFO", "DEBUG", "NOTSET"]

# Flask
try:
    from petisco.frameworks.flask.application.flask_application import FlaskApplication
    flask = ["FlaskApplication"]
except (RuntimeError, ImportError):
    flask = ""

__all__ = classes + constants + flask



from typing import List

from meiga import Result
from meiga.decorators import meiga

from petisco.logger.logger import ERROR, INFO
from petisco.logger.log_message import LogMessage
from petisco.use_case.use_case import UseCase


class UseCaseLogger(object):
    def __init__(self, logger=None, logging_parameters_whitelist: List[str] = None):
        self.logger = logger
        self.logging_parameters_whitelist = logging_parameters_whitelist

    def __call__(self, cls):
        if not issubclass(cls, UseCase):
            raise TypeError("UseCaseLogger only decorates UseCase subclasses")

        class UseCaseWrapped(cls):
            logger = self.logger
            logging_parameters_whitelist = self.logging_parameters_whitelist

            def execute(self, *args, **kwargs):
                correlation_id = kwargs.get("correlation_id")

                log_message = LogMessage(
                    layer="use_case",
                    operation=f"{cls.__name__}",
                    correlation_id=correlation_id,
                )

                log_message.message = f"Start"
                self.logger.log(INFO, log_message.to_json())

                if self.logging_parameters_whitelist:
                    loggable_kwargs = [
                        (k, v)
                        for k, v in kwargs.items()
                        if k in self.logging_parameters_whitelist
                    ]

                    if loggable_kwargs:
                        log_message.message = dict(loggable_kwargs)
                        self.logger.log(INFO, log_message.to_json())

                result = self._run_execute(*args, **kwargs)

                if not isinstance(result, Result):
                    return result

                detail = getattr(result.value, "message", None)
                if detail:
                    detail = f"-> {detail}"
                else:
                    detail = ""

                if result.is_failure:
                    log_message.message = f"{result} {detail}"
                    self.logger.log(ERROR, log_message.to_json())
                else:
                    log_message.message = f"{result.value}"
                    self.logger.log(INFO, log_message.to_json())

                return result

            @meiga
            def _run_execute(self, *args, **kwargs) -> Result:
                return super().execute(*args, **kwargs)

        return UseCaseWrapped


use_case_logger = UseCaseLogger
